/*
SCARF Association Rule Mining Program
Copyright (c) 2012-13 Balazs Szalkai

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "bestrules.hpp"

bool BestRules::shouldInsertRule(const double goodness) const
{
    return items().size() == 0 || items().back().stats.goodness() < goodness;
}

Nullable<size_t> BestRules::getRuleInsertIndex(const double goodness) const
{
    const size_t n = items().size();
    // better than the best rule?
    if (n == 0 || at(items(), 0).stats.goodness() < goodness) return 0;
    // worse than the worst best rule?
    if (items().back().stats.goodness() >= goodness)
    {
        // rule table is full --> discard
        if (n == f_capacity) return Nullable<size_t>();
        // otherwise --> register
        return n;
    }

    // lo: the index of a rule which is definitely better or equal to the given rule
    size_t lo = 0;
    // hi: the index of a rule which is definitely worse than the given rule
    size_t hi = n-1;

    while (hi-lo > 1)
    {
        size_t mid = (lo+hi)/2;
        if (at(items(), mid).stats.goodness() < goodness) hi = mid; else lo = mid;
    }

    return hi;
}

void BestRules::insertRule(const Rule &rule)
{
    Nullable<size_t> insertIndex = getRuleInsertIndex(rule.stats.goodness());
    if (insertIndex.isNull()) return;

    std::int64_t hash = rule.getHash();
    if (f_hashes.find(hash) != f_hashes.end()) return;

    f_items.insert(f_items.begin() + (size_t)insertIndex, rule);
    f_hashes.insert(hash);
    if (f_items.size() > f_capacity)
    {
        f_hashes.erase(f_items.back().getHash());
        f_items.pop_back();
    }
}

void BestRules::merge(const BestRules& other)
{
    for (const Rule& r : other.items()) insertRule(r);
}
