/*
SCARF Association Rule Mining Program
Copyright (c) 2012-13 Balazs Szalkai

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#pragma once

#include <cstddef>
#include <cstdint>
#include "utils.hpp"
#include <vector>

template <class Bits>
Bits Zeros()
{
    return Bits(0);
}

template <class Bits>
Bits Ones()
{
    return ~Zeros<Bits>();
}

/// Gets a bit from a vector storing unsigned integers.
template <class T>
bool getBit(const std::vector<T>& v, size_t index)
{
    constexpr size_t bits = sizeof(T)*8;
    typename std::vector<T>::const_reference x = at(v, index / bits);
    T bit = (T(1) << (index % bits));

    return x & bit;
}

/// Clears or sets a bit in a vector storing unsigned integers.
template <class T>
void setBit(std::vector<T>& v, size_t index, bool value)
{
    constexpr size_t bits = sizeof(T)*8;
    typename std::vector<T>::reference x = at(v, index / bits);
    T bit = (T(1) << (index % bits));

    if (value) x |= bit; else x &= ~bit;
}

class SimplePopCount
{
    /// Returns the number of set bits.
    /// T must be an unsigned integer type.
    public: template <class T>
    static size_t get(T x)
    {
        size_t result = 0;
        while (x)
        {
            x &= x-1;
            ++result;
        }
        return result;
    }
};

class LookupPopCount
{
    private: static char popCountTable[];

    /// Must be called at least once before using this class.
    public: static void init();

    public: static size_t get(uint16_t x);
    public: static size_t get(uint32_t x);
    public: static size_t get(uint64_t x);
};

/// Dark magic holds this together. See http://en.wikipedia.org/wiki/Hamming_weight.
class AdvancedPopCount
{
    public: static size_t get(uint32_t);
    public: static size_t get(uint64_t);
};

class DefaultPopCount
{
    /// Must be called at least once before using this class.
    public: static void init();

    public: template <class T>
    static size_t get(const T x)
    {
        return LookupPopCount::get(x); // this was the fastest according to the tests
    }
};
