/*
SCARF Association Rule Mining Program
Copyright (c) 2012-13 Balazs Szalkai

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#pragma once

#include <cstddef>
#include <queue>
#include "tinythread.h"
#include <vector>

//#include <iostream>

template <class Input, class Output, class Worker>
class JobPool
{
    private: const Worker& worker;

    // This mutex should be used when accessing the objects in this group.
    private: tthread::mutex mutex;
    private: std::queue<Input> jobs;
    private: std::queue<Output> results;
    private: size_t jobsPending;

    private: bool threadsShouldStop;
    private: std::vector<tthread::thread*> workerThreads;
    private: tthread::condition_variable jobAvailableCondition;
    private: tthread::condition_variable resultReadyCondition;

    public: JobPool(const Worker& _worker, const size_t nThreads)
        : worker(_worker),
            jobsPending(0),
            threadsShouldStop(false)
    {
        tthread::lock_guard<tthread::mutex> lock(mutex);

//        std::cerr<<"creating pool"<<std::endl;
        workerThreads.reserve(nThreads);
        for (size_t i = 0; i < nThreads; ++i)
            workerThreads.push_back(new tthread::thread(staticThreadFunction, static_cast<void*>(this)));
    }

    public: JobPool(const JobPool&) = delete;
    public: JobPool(JobPool&& other) = delete;
    public: JobPool& operator=(const JobPool&) = delete;
    public: JobPool& operator=(JobPool&&) = delete;

    public: ~JobPool()
    {
        {
            tthread::lock_guard<tthread::mutex> lock(mutex);

            threadsShouldStop = true;
            // wake up all possibly waiting threads
            for (size_t i = 0; i < workerThreads.size(); ++i)
                jobAvailableCondition.notify_one();
        }

        for (tthread::thread* t : workerThreads)
        {
            t->join();
            delete t;
        }
//        std::cerr<<"pool destroyed"<<std::endl;
    }

    public: void push(const Input& input)
    {
//        std::cerr<<"pushed input (copy)"<<std::endl;
        tthread::lock_guard<tthread::mutex> lock(mutex);

        jobs.push(input);
        ++jobsPending;
        jobAvailableCondition.notify_one();
    }

    public: void push(Input&& input)
    {
//        std::cerr<<"pushed input (move)"<<std::endl;
        tthread::lock_guard<tthread::mutex> lock(mutex);

        jobs.push(std::move(input));
        ++jobsPending;
        jobAvailableCondition.notify_one();
    }

    /// Pops a job result.
    /// Waits for one if there are pending jobs.
    /// If no jobs pending and no results available, returns false.
    public: bool waitAndPop(Output& dest)
    {
        tthread::lock_guard<tthread::mutex> lock(mutex);

        if (jobsPending == 0)
        {
//            std::cerr<<"no jobs pending, returning false"<<std::endl;
            return false;
        }
        --jobsPending; // reserve one

        while (results.empty())
            resultReadyCondition.wait(mutex);
        dest = std::move(results.front());
        results.pop();
//        std::cerr<<"popped a result"<<std::endl;
        return true;
    }

    /// TinyThread hack.
    private: static void staticThreadFunction(void* self)
    {
        static_cast<JobPool*>(self)->threadFunction();
    }

    /// This is the thread function. Do not call directly.
    private: void threadFunction()
    {
//        std::cerr<<"thread start"<<std::endl;
        while (!threadsShouldStop)
        {
            Input input;

            {
                tthread::lock_guard<tthread::mutex> lock(mutex);
                while (!threadsShouldStop && jobs.empty())
                    jobAvailableCondition.wait(mutex);
                if (threadsShouldStop) break;
                input = std::move(jobs.front());
                jobs.pop();
            }

            Output output = worker(input); // do the calculation

            {
                tthread::lock_guard<tthread::mutex> lock(mutex);
                results.push(std::move(output));
                resultReadyCondition.notify_one();
            }
        }
//        std::cerr<<"thread stop"<<std::endl;
    }
};
