/*
SCARF Association Rule Mining Program
Copyright (c) 2012-13 Balazs Szalkai

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#pragma once

#include "bits.hpp"
#include "rule.hpp"
#include <set>
#include <string>
#include <vector>

class ConjunctiveNormalForm;
class Params;
class Table;

/// A column descriptor for a Mine object.
class MineColumn
{
    /// Bits are ordered into blocks in a mine column.
    /// This is the type of a block. It must be an unsigned integral type.
    public: typedef uintptr_t block_t;

	/// The name for the column.
	public: std::string name;
	/// The index of the column group (an index to Mine::columnGroups).
	public: size_t groupIndex;
	/// Whether the column is allowed to occur on the RHS of a rule.
    public: bool rhsAllowed;

    // Data for this column. Stored as bit array.
    public: std::vector<block_t> dataTrue; ///< does the cell contain 'true'?
    public: std::vector<block_t> dataNotNA; ///< does the cell contain a non-NA value?

    // Statistics for the column
	/// The number of rows containing non-NA values for this column.
	public: size_t universe;
	/// The number of rows containing 'true' for this column.
	public: size_t support;

	public: MineColumn(const std::string &_name)
        : name(_name), groupIndex(0), rhsAllowed(false),
        universe(0), support(0)
	{
	}

	public: class DontCopyData {};
    /// Copy constructor which does not copy the data itself
	public: MineColumn(const MineColumn& other, DontCopyData)
        : name(other.name), groupIndex(other.groupIndex), rhsAllowed(other.rhsAllowed),
        universe(0), support(0)
	{
	}

    /// Copies those rows from another column whose indices are contained by keepRows.
	public: MineColumn(const MineColumn& other, const std::set<size_t>& keepRows);
    /// Loads the column data from a table and calculates statistics.
	public: void load(const Table& table, size_t columnIndex, const std::string& valueSet);

	private: void calculateTransientValues();

    /// Each mine column should have the same number of rows.
	public: size_t rowCount() const
	{
	    return dataTrue.size() * 8*sizeof(block_t);
	}

	public: NullableBool getCell(size_t rowIndex) const
	{
        if (!getBit(dataNotNA, rowIndex)) return NullableBool::Null;
        return getBit(dataTrue, rowIndex) ? NullableBool::True : NullableBool::False;
	}

	public: void setCell(size_t rowIndex, NullableBool value)
	{
	    setBit(dataTrue, rowIndex, value == NullableBool::True);
	    setBit(dataNotNA, rowIndex, value != NullableBool::Null);
	}
};

/// If a column of a group occurs on the RHS, then other columns in that group cannot occur on LHS.
class ColumnGroup
{
    /// The maximal total number of occurrences on the LHS of all the columns belonging to this group
    public: size_t maxLhsCount;
};

/// A Mine is a table where each cell may contain only true, false or NA.
class Mine
{
    protected: std::vector<MineColumn> f_columns;
    /// Database columns
    public: const std::vector<MineColumn>& columns() const
    {
        return f_columns;
    }

	protected: std::vector<ColumnGroup> f_columnGroups;
    /// Group descriptors.
	public: const std::vector<ColumnGroup>& columnGroups() const
	{
	    return f_columnGroups;
	}

    /// Creates an empty mine.
    public: Mine() {}

    /// Prepares a data table for mining.
    public: explicit Mine(const Table& table, const Params& params);
    /// Copies those rows from another mine whose indices are contained by keepRows.
    public: explicit Mine(const Mine& other, const std::set<size_t>& keepRows);

    /// Gets the number of rows in the mine.
    public: size_t rowCount() const;

    /// Tests a rule and calculates its statistics.
    public: void getRuleStats(Rule& rule, const double eValueFactor) const;

    /// Tests a rule and calculates its statistics.
    /// It intersects the universe of the rule with the universe for the elements of universeColumns.
    public: void getRuleStats(Rule& rule, const double eValueFactor,
        const std::vector<size_t>* const universeColumns) const;
};
