/*
SCARF Association Rule Mining Program
Copyright (c) 2012-13 Balazs Szalkai

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "miner.hpp"

#include <algorithm>
#include <iostream>

/// WARNING: This class has reference fields.
/// A leaf (elementary clause) of the mining pattern.
class Leaf
{
    /// The column index in a CNF.
    public: Nullable<size_t>& content;
    /// The index of 'content' in the 'lhsIndexes' vector.
    public: Nullable<size_t> lhsIndexes_index;
    // Milestone: the end of an OR clause
    public: const bool isMilestone;
    // We will only mine such patterns where leaves are ordered within an OR clause by index
    // and OR clauses with the same length are ordered by the index of the value of the starting leaf.
    // Thus, generally there will be a lower bound for the value of a leaf, namely the value of another leaf.
    // This field stores the index of that leaf. Can be null (not used).
    public: Nullable<size_t> lowerBoundIndex;

    public: Leaf(Nullable<size_t>& _content, const bool _isMilestone)
        : content(_content), lhsIndexes_index(),
        isMilestone(_isMilestone), lowerBoundIndex()
    {
    }
};

std::vector<size_t> RuleMiner::get_lhsColumnIndexes() const
{
    std::vector<size_t> result;
    for (size_t i = 0; i < mine.columns().size(); i++)
    {
        const MineColumn &col = at(mine.columns(), i);

        if (col.universe >= params.minStats.universe && col.support > 0 &&
            at(mine.columnGroups(), col.groupIndex).maxLhsCount > 0)
                result.push_back(i);
    }
    return result;
}

std::vector<size_t> RuleMiner::get_rhsColumnIndexes() const
{
    std::vector<size_t> result;
    for (size_t i = 0; i < mine.columns().size(); ++i)
    {
        const MineColumn &col = at(mine.columns(), i);

        if (col.universe >= params.minStats.universe && col.rhsAllowed &&
            col.support >= std::max(params.minStats.support, params.minStats.supportRhs))
                result.push_back(i);
    }
    return result;
}

/// WARNING: This class has reference fields.
class RuleMinerAction
{
    // INPUT
    public: Mine mine; // the algorithm requires copying the mine
    public: const Params& params;
	public: const ConcurrentStream& logger;
	public: const double eValueFactor;
    // OUTPUT
    public: BestRules bestRules;

    // Temporary fields for computation
    private: Rule rule;
    private: std::vector<Leaf> leaves;
    private: std::vector<size_t> lhsIndexes;
    /// The current count of lhs leaves per group
    private: std::vector<size_t> currentGroupSize;

    public: RuleMinerAction(const Mine& _mine,
                       const Params& _params,
					   const ConcurrentStream& _logger,
                       const size_t _rhsIndex,
                       const double _eValueFactor)
        : mine(), params(_params), logger(_logger), eValueFactor(_eValueFactor),
        bestRules(params.maxRuleCount)
    {
        rule.rhsColumnIndex = _rhsIndex;
        rule.lhs.parsePattern(params.patternStr, true);

        // Make leaves
        size_t prevLength = 0; // length of previous children
        for (const Range<size_t>& range : rule.lhs.getPattern())
        {
            for (size_t i : range)
            {
                Leaf leaf(at(rule.lhs.nodes, i).column, i == range.back());

                // lower bound:
                // * previous node in the same OR clause?
                if (i > range.front()) leaf.lowerBoundIndex = leaves.size() - 1;
                // * starting node of prev. OR clause, in some cases?
                else if (prevLength == range.size())
                    leaf.lowerBoundIndex = leaves.size() - prevLength;

                leaves.push_back(leaf);
            }

            prevLength = range.size();
        }

        // Prepare the mine
        // Delete those mine rows where RHS is N/A
        {
            std::set<size_t> keepRows;
            const MineColumn& rhsColumn = at(_mine.columns(), rule.rhsColumnIndex);
            for (size_t i = 0, n = _mine.rowCount(); i < n; ++i)
                if (rhsColumn.getCell(i) != NullableBool::Null)
                    keepRows.insert(i);
            mine = Mine(_mine, keepRows);
        }

        // collect columns which can appear on LHS
        lhsIndexes = RuleMiner(mine, params, logger).get_lhsColumnIndexes();

        // check: selected column cannot appear on RHS?
        if (!at(mine.columns(), rule.rhsColumnIndex).rhsAllowed)
            throw std::invalid_argument("Column cannot appear on RHS");

        // Sort LHS clauses by support ascending
        // This greatly improves speed of search for association rules (easy to see)!
        // The branches of the search tree are cut off earlier using this approach
        logger<<"Number of LHS columns: "<<lhsIndexes.size()<<std::endl;
        std::sort(lhsIndexes.begin(), lhsIndexes.end(),
            [&](const size_t a, const size_t b) { return at(mine.columns(), a).support < at(mine.columns(), b).support; });

        currentGroupSize.resize(mine.columnGroups().size(), 0);
        // don't allow RHS column to appear on LHS
        at(currentGroupSize, at(mine.columns(), rule.rhsColumnIndex).groupIndex) = std::numeric_limits<size_t>::max();

        recurse(0);
    }

    private: void recurse(size_t leafIndex)
    {
        Leaf &leaf = at(leaves, leafIndex);
        size_t start = 0;
        if (!leaf.lowerBoundIndex.isNull())
        {
            const Nullable<size_t>& bound = at(leaves, (size_t)leaf.lowerBoundIndex).lhsIndexes_index;
            start = 1 + (size_t)bound; // bound is never null
        }

        // try all values for lhs leaf
        for (size_t i = start; i < lhsIndexes.size(); ++i)
        {
            const size_t lhsColumn = at(lhsIndexes, i);
            const MineColumn& col = at(mine.columns(), lhsColumn);

            if (at(currentGroupSize, col.groupIndex) < at(mine.columnGroups(), col.groupIndex).maxLhsCount)
            {
                ++at(currentGroupSize, col.groupIndex);

                leaf.content = lhsColumn;
                leaf.lhsIndexes_index = i;

    /*            for (const Leaf& leaf : leaves)
                    logger<<leaf.content<<' ';
                logger<<std::endl;
                logger.flush();*/

                if (leafIndex == 0 && i % 5 == 0)
                    logger<<i<<' ';

                mine.getRuleStats(rule, eValueFactor);

                // too small support and it cannot increase -> don't recurse!
                bool canRecurse = !(leaf.isMilestone &&
                    (rule.stats.universe < params.minStats.universe ||
                     rule.stats.support < params.minStats.support ||
                     rule.stats.supportLhs < params.minStats.supportLhs));
                if (canRecurse)
                {
                    // found a good rule?
                    if (rule.stats.atLeastAsGoodAs(params.minStats) &&
                        bestRules.shouldInsertRule(rule.stats.goodness()) &&
                        !rule.canSimplify(mine, params.minStats, eValueFactor, params.allowedGoodnessDecrease))
                            bestRules.insertRule(rule);

                    if (leafIndex+1 < leaves.size()) recurse(leafIndex+1);
                }

                --at(currentGroupSize, col.groupIndex);
            }
        }

        leaf.content.setNull();
        leaf.lhsIndexes_index.setNull();
    }
};

RuleMiner::RuleMiner(const Mine& _mine, const Params& _params, const ConcurrentStream& _logger)
    : mine(_mine), params(_params), logger(_logger)
{
    ConjunctiveNormalForm cnf;
    cnf.parsePattern(params.patternStr, false);
    eValueFactor = std::pow(get_lhsColumnIndexes().size(), cnf.nodes.size()) * get_rhsColumnIndexes().size();
}

BestRules RuleMiner::operator()(size_t rhsIndex) const
{
    const MineColumn& rhsColumn = at(mine.columns(), rhsIndex);
    logger<<rhsColumn.name<<":"<<std::endl; // notification

    RuleMinerAction rma(mine, params, logger, rhsIndex, eValueFactor);
    return std::move(rma.bestRules);
}
