/*
SCARF Association Rule Mining Program
Copyright (c) 2012-13 Balazs Szalkai

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "params.hpp"

#include <algorithm>
#include <cctype>
#include <fstream>
#include <iostream>
#include "pathutils.hpp"
#include <stdexcept>
#include "utils.hpp"

std::vector<std::string> AllValueSets::getValueSets(const std::string& values)
{
	std::vector<std::string> result;
	const size_t n = values.size();
    if (n <= 1) return result;
    if (n >= 10)
        throw std::invalid_argument("Error: too many value sets when using ValueSets::All.");

    result.reserve((1u << n) - 2);
    for (size_t i = 1, iMax = (1u << n) - 1; i < iMax; ++i)
    {
        std::string s;
        for (size_t j = 0; j < n; ++j)
            if ((i & (1u << j)) != 0) s += at(values, j);
        result.push_back(s);
    }
    return result;
}

int RangeValueSets::compareValues(char c1, char c2)
{
    size_t index1 = ordering.find(c1), index2 = ordering.find(c2);
    if (index1 == std::string::npos)
        return index2 == std::string::npos ? compare(c1, c2) : -1;
    return index2 == std::string::npos ? 1 : compare(index1, index2);
}

std::vector<std::string> RangeValueSets::getValueSets(const std::string& values)
{
	std::vector<std::string> result;
	const size_t n = values.size();
    if (n <= 1) return result;

    // sort values
    std::string sorted = values;
    std::sort(sorted.begin(), sorted.end(),
        [&](const char a, const char b){return compareValues(a, b) < 0;});
    result.reserve(2*(sorted.size()-1)); // here we use that n > 0 !
    for (auto it = sorted.begin(); it != sorted.end(); ++it)
    {
        if (it != sorted.begin())
            result.emplace_back(it, sorted.end());
        if (it+1 != sorted.end())
            result.emplace_back(sorted.begin(), it+1);
    }
    return result;
}

std::vector<std::string> SingleValueSets::getValueSets(const std::string& values)
{
	std::vector<std::string> result;
	const size_t n = values.size();
    if (n <= 1) return result;

    result.reserve(n);
    for (char c : values) result.emplace_back(1, c);
    return result;
}

ColumnParams::ColumnParams(const std::string& s, const ConcurrentStream& logger)
    : ColumnParams()
{
    const std::vector<std::string> tokens = split(s, isWhitespace, EmptyEntryAction::Remove);
    if (tokens.size() < 3) return;

    name = at(tokens, 0);

    {
        const std::string vs = at(tokens, 1);
        if (vs == "all") valueSets = std::unique_ptr<IValueSets>(new AllValueSets); else
        if (startsWith(vs, const_string<char>("range")))
        {
            std::string ordering;
            const_string<char> prefix = "range:";
            if (startsWith(vs, prefix))
                ordering = vs.substr(prefix.size());
            valueSets = std::unique_ptr<IValueSets>(new RangeValueSets(ordering));
        } else
        if (vs == "single") valueSets = std::unique_ptr<IValueSets>(new SingleValueSets); else
        {
            logger<<"Error: Invalid value set specification: "<<vs<<std::endl;
            throw std::invalid_argument("Invalid value set: "+vs);
        }
    }

	const std::string& side = at(tokens, 2);
    lhs = (side.find("lhs") != std::string::npos);
    rhs = (side.find("rhs") != std::string::npos);
}

Params::Params(const std::string &fn, const ConcurrentStream& logger)
    : nThreads(0), maxRuleCount(100), allowedGoodnessDecrease(0.98)
{
    const std::string filePath = extractPath(fn);
    std::ifstream f(fn);

    while (f)
    {
        std::string key;
        std::getline(f, key);
        if (key.size() == 0) continue;
        std::transform(key.begin(), key.end(), key.begin(),
            [](const char a){return std::tolower(a);});

        std::string value;
        std::getline(f, value);

        switch (getHash(key))
        {
            case getHash("table"): tableFile = combinePath(filePath, value); break;
            case getHash("output"):
                outputFile = value.size()>0 ? combinePath(filePath, value) : std::string(); break;
            case getHash("xml_output"): xmlOutputFile = combinePath(filePath, value); break;
            case getHash("nthreads"): nThreads = parseInt(value); break;
            case getHash("pattern"): patternStr = value; break;
            case getHash("rulecount"): maxRuleCount = parseInt(value); break;

            // minStats
            case getHash("min_universe"): minStats.universe = parseInt(value); break;
            case getHash("min_lhs_support"): minStats.supportLhs = parseInt(value); break;
            case getHash("min_rhs_support"): minStats.supportRhs = parseInt(value); break;
            case getHash("min_support"): minStats.support = parseInt(value); break;
            case getHash("min_confidence"): minStats.confidence = std::atof(value.c_str()); break;
            case getHash("min_lift"): minStats.lift = std::atof(value.c_str()); break;
            case getHash("min_leverage"): minStats.leverage = std::atof(value.c_str()); break;
            case getHash("max_pvalue"): minStats.pValue = std::atof(value.c_str()); break;

            // allowed goodness decrease when simplifying
            case getHash("simpl_param"):
                allowedGoodnessDecrease = std::atof(value.c_str()); break;

            // column
            case getHash("column"): {
                ColumnParams p(value, logger);
                // p will be moved, so store name here to prevent lifetime issues
                const std::string columnName = p.name;
                columnParams[columnName] = std::move(p);
            } break;

            default:
                logger<<"Warning: Unparsed key in parameter file: "<<key<<std::endl;
                break;
        }
    }
}
