/*
SCARF Association Rule Mining Program
Copyright (c) 2012-13 Balazs Szalkai

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "rulestats.hpp"

#include "chisquared.hpp"
#include "utils.hpp"

void RuleStats::calculateTransientValues(const double eValueFactor)
{
    const double u = 1.0/universe;
    const double p = support*u, px = supportLhs*u, py = supportRhs*u;

    confidence = p / px;
    lift = p / (px*py);
    leverage = p - px*py;

    // Formula for 2x2 chi-squared statistic calculation:
    // X2 = det^2 * universe / (supp(x)*supp(not x)*supp(y)*supp(not y))
    double chiSquared = sqr(px*py - (1-px)*(1-py)) / (px*(1-px)*py*(1-py));
    pValue = ChiSquared::pValue(1.0, chiSquared);
    eValue = eValueFactor * pValue;
}

bool RuleStats::atLeastAsGoodAs(const RuleStats &other) const
{
    return universe >= other.universe &&
        supportLhs >= other.supportLhs &&
        supportRhs >= other.supportRhs &&
        support >= other.support &&
        confidence >= other.confidence &&
        lift >= other.lift &&
        leverage >= other.leverage &&
		pValue <= other.pValue &&
		eValue <= other.eValue;
}

std::string RuleStats::toString() const
{
    return "Universe: "+::toString(universe)+
        ", LHS support: "+::toString(supportLhs)+
        ", RHS support: "+::toString(supportRhs)+
        ", Support: "+::toString(support)+

        "\nConfidence: "+::toString(confidence)+
        ", Lift: "+::toString(lift)+
        ", Leverage: "+::toString(leverage)+
        ", p-value: "+::toString(pValue)+
        ", E-value: "+::toString(eValue)+

        "\n"+::toString(goodness())+"\n";
}

void RuleStats::toXml(XmlWriter& xml) const
{
    xml.start("stats");
    xml.start("universe"); xml.text(::toString(universe)); xml.end();
    xml.start("lhs_support"); xml.text(::toString(supportLhs)); xml.end();
    xml.start("rhs_support"); xml.text(::toString(supportRhs)); xml.end();
    xml.start("support"); xml.text(::toString(support)); xml.end();
    xml.start("confidence"); xml.text(::toString(confidence)); xml.end();
    xml.start("lift"); xml.text(::toString(lift)); xml.end();
    xml.start("leverage"); xml.text(::toString(leverage)); xml.end();
    xml.start("p_value"); xml.text(::toString(pValue)); xml.end();
    xml.start("e_value"); xml.text(::toString(eValue)); xml.end();
    xml.end();
}
