/*
SCARF Association Rule Mining Program
Copyright (c) 2012-13 Balazs Szalkai

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#pragma once

#include "logicalexpr.hpp"
#include "params.hpp"
#include <set>
#include <string>
#include "utils.hpp"
#include <vector>

template<class Column, class Row>
class AbstractTable
{
	protected: std::vector<Column> f_columns;
	protected: std::vector<Row> f_rows;

	public: const std::vector<Column>& columns;
	public: const std::vector<Row>& rows;

    protected: AbstractTable()
        : columns(f_columns), rows(f_rows)
    {
    }

    protected: virtual ~AbstractTable() {}

	protected: void clear()
	{
		f_columns.clear();
		f_rows.clear();
	}
};

/// A column descriptor for a Table object.
class TableColumn
{
	/// The name for the column.
	public: std::string name;
	/// The number of rows containing non-NA values for this column.
	public: size_t universe;
	/// The value set of this column.
	public: std::set<char> values;
	/// Were there cells with length at least 2 which had to be trimmed to the first char?
	public: bool hadLongCells;

	public: TableColumn(const std::string &_name)
        : name(_name), universe(0), hadLongCells(false)
	{
	}
};

/// Stores a table whose cells are all one character wide.
/// Each row is a string whose length is columnCount()
class Table : virtual public AbstractTable<TableColumn, std::string>
{
    /// Creates an empty table.
    public: Table() {}

	/// Loads a CSV file as a table.
	/// The first line of the file must contain the header (column names).
	/// The subsequent lines must contain the data.
	/// Only the first character of each data cell is read.
	/// Missing data (N/A) must be denoted with an empty cell.
	public: void load(const std::string &fileName);

	public: static bool validColumnName(const std::string& name);
};
